@@ -42,29 +42,50 @@ module Qt
 	def Qt.debug_level
 		@@debug_level
 	end
-		
+	
+	module Internal
+		#
+		# From the enum MethodFlags in qt-copy/src/tools/moc/generator.cpp
+		#
+		AccessPrivate = 0x00
+		AccessProtected = 0x01
+		AccessPublic = 0x02
+		MethodMethod = 0x00
+		MethodSignal = 0x04
+		MethodSlot = 0x08
+		MethodCompatibility = 0x10
+		MethodCloned = 0x20
+		MethodScriptable = 0x40
+	end
+	
 	class Base
 		def self.signals(*signal_list)
 			meta = Qt::Meta[self.name] || Qt::MetaInfo.new(self)
-			meta.add_signals(signal_list)
+			meta.add_signals(signal_list, Internal::MethodSignal | Internal::AccessProtected)
 			meta.changed = true
 		end
 	
 		def self.slots(*slot_list)
 			meta = Qt::Meta[self.name] || Qt::MetaInfo.new(self)
-			meta.add_slots(slot_list)
+			meta.add_slots(slot_list, Internal::MethodSlot | Internal::AccessPublic)
+			meta.changed = true
+		end
+
+		def self.private_slots(*slot_list)
+			meta = Qt::Meta[self.name] || Qt::MetaInfo.new(self)
+			meta.add_slots(slot_list, Internal::MethodSlot | Internal::AccessPrivate)
 			meta.changed = true
 		end
 
 		def self.q_signal(signal)
 			meta = Qt::Meta[self.name] || Qt::MetaInfo.new(self)
-			meta.add_signals([signal])
+			meta.add_signals([signal], Internal::MethodSignal | Internal::AccessProtected)
 			meta.changed = true
 		end
 
 		def self.q_slot(slot)
 			meta = Qt::Meta[self.name] || Qt::MetaInfo.new(self)
-			meta.add_slots([slot])
+			meta.add_slots([slot], Internal::MethodSlot | Internal::AccessPublic)
 			meta.changed = true
 		end
 
@@ -180,6 +201,13 @@ module Qt
 			return [self] + ids.map {|id| Qt::Internal.find_class(Qt::Internal.classid2name(id))} + klasses
 		end
 
+		# Change the behaviors of is_a? and kind_of? (alias of is_a?) to use above self.ancestors method
+		# Note: this definition also affects Object#===
+		def is_a?(mod)
+			super || self.class.ancestors.include?(mod)
+		end
+		alias :kind_of? :is_a?
+
 		def methods(regular=true)
 			if !regular
 				return singleton_methods
@@ -188,16 +216,16 @@ module Qt
 			qt_methods(super, 0x0)
 		end
 	
-		def protected_methods
+		def protected_methods(all=true)
 			# From smoke.h, Smoke::mf_protected 0x80
 			qt_methods(super, 0x80)
 		end
 	
-		def public_methods
+		def public_methods(all=true)
 			methods
 		end
 	
-		def singleton_methods
+		def singleton_methods(all=true)
 			# From smoke.h, Smoke::mf_static 0x01
 			qt_methods(super, 0x01)
 		end
@@ -680,32 +708,60 @@ module Qt
 		def serviceOwner(name)
     		return Qt::DBusReply.new(internalConstCall(Qt::DBus::AutoDetect, "GetNameOwner", [Qt::Variant.new(name)]))
 		end
+		
+		def service_owner(name)
+    		return serviceOwner(name)
+		end
 
 		def registeredServiceNames
 			return Qt::DBusReply.new(internalConstCall(Qt::DBus::AutoDetect, "ListNames"))
 		end
 
+		def registered_service_names
+			return registeredServiceNames
+		end
+
 		def isServiceRegistered(serviceName)
     		return Qt::DBusReply.new(internalConstCall(Qt::DBus::AutoDetect, "NameHasOwner", [Qt::Variant.new(serviceName)]))
 		end
 
+		def is_service_registered(serviceName)
+    		return isServiceRegistered(serviceName)
+		end
+
 		def serviceRegistered?(serviceName)
     		return isServiceRegistered(serviceName)
 		end
 
+		def service_registered?(serviceName)
+    		return isServiceRegistered(serviceName)
+		end
+
 		def servicePid(serviceName)
     		return Qt::DBusReply.new(internalConstCall(Qt::DBus::AutoDetect, "GetConnectionUnixProcessID", [Qt::Variant.new(serviceName)]))
 		end
 
+		def service_pid(serviceName)
+    		return servicePid(serviceName)
+		end
+
 		def serviceUid(serviceName)
     		return Qt::DBusReply.new(internalConstCall(Qt::DBus::AutoDetect, "GetConnectionUnixUser", [Qt::Variant.new(serviceName)]))
 		end
 
+		def service_uid(serviceName)
+    		return serviceUid(serviceName)
+		end
+
 		def startService(name)
     		return call("StartServiceByName", Qt::Variant.new(name), Qt::Variant.new(0)).value
 		end
-	end
 
+		def start_service(name)
+    		startService(name)
+		end
+	end
+	
 	class DBusError < Qt::Base 
 		def type(*args)
 			method_missing(:type, *args)
@@ -716,6 +772,10 @@ module Qt
 		def call(method_name, *args)
 			if args.length == 0
 				return super(method_name)
+			elsif method_name.is_a? Qt::Enum
+				opt = args.shift
+				qdbusArgs = args.collect {|arg| qVariantFromValue(arg)}
+				return super(method_name, opt, *qdbusArgs)
 			else
 				# If the method is Qt::DBusInterface.call(), create an Array 
 				# 'dbusArgs' of Qt::Variants from '*args'
@@ -1039,6 +1099,12 @@ module Qt
 		end
 	end
 
+	class GraphicsSceneDragDropEvent < Qt::Base 
+		def type(*args)
+			method_missing(:type, *args)
+		end
+	end
+
 	class GraphicsSceneMouseEvent < Qt::Base 
 		def type(*args)
 			method_missing(:type, *args)
@@ -1202,6 +1268,15 @@ module Qt
 	end
 
 	class ItemSelection < Qt::Base
+		include Enumerable
+
+		def each
+			for i in 0...count
+				yield at(i)
+			end
+			return self
+		end
+
 		def select(*args)
 			method_missing(:select, *args)
 		end
@@ -1256,7 +1331,7 @@ module Qt
 
 	class ListWidgetItem < Qt::Base
 		def clone(*args)
-			method_missing(:clone, *args)
+			Qt::ListWidgetItem.new(self)
 		end
 
 		def type(*args)
@@ -1265,12 +1340,12 @@ module Qt
 
 		def inspect
 			str = super
-			str.sub(/>$/, " text=%s>" % text)
+			str.sub(/>$/, " text='%s'>" % text)
 		end
 		
 		def pretty_print(pp)
 			str = to_s
-			pp.text str.sub(/>$/, " text=%s>" % text)
+			pp.text str.sub(/>$/, " text='%s'>" % text)
 		end
 	end
 
@@ -1855,9 +1930,23 @@ module Qt
 	end
 
 	class StandardItem < Qt::Base 
+		def inspect
+			str = super
+			str.sub(/>$/, " text='%s'>" % [text])
+		end
+		
+		def pretty_print(pp)
+			str = to_s
+			pp.text str.sub(/>$/, "\n text='%s'>" % [text])
+		end
+
 		def type(*args)
 			method_missing(:type, *args)
 		end
+
+		def clone
+			Qt::StandardItem.new(self)
+		end
 	end
 
 	class StandardItemModel < Qt::Base 
@@ -1892,7 +1981,7 @@ module Qt
 
 	class TableWidgetItem < Qt::Base
 		def clone(*args)
-			method_missing(:clone, *args)
+			Qt::TableWidgetItem.new(self)
 		end
 
 		def type(*args)
@@ -1901,12 +1990,12 @@ module Qt
 
 		def inspect
 			str = super
-			str.sub(/>$/, " text=%s>" % text)
+			str.sub(/>$/, " text='%s'>" % text)
 		end
 		
 		def pretty_print(pp)
 			str = to_s
-			pp.text str.sub(/>$/, " text=%s>" % text)
+			pp.text str.sub(/>$/, " text='%s'>" % text)
 		end
 	end
 
@@ -2069,7 +2158,7 @@ module Qt
 			str.sub!(/>$/, "")
 			str << " parent=%s," % parent unless parent.nil?
 			for i in 0..(columnCount - 1)
-				str << " text%d=%s," % [i, self.text(i)]
+				str << " text%d='%s'," % [i, self.text(i)]
 			end
 			str.sub!(/,?$/, ">")
 		end
@@ -2079,14 +2168,14 @@ module Qt
 			str.sub!(/>$/, "")
 			str << " parent=%s," % parent unless parent.nil?
 			for i in 0..(columnCount - 1)
-				str << " text%d=%s," % [i, self.text(i)]
+				str << " text%d='%s'," % [i, self.text(i)]
 			end
 			str.sub!(/,?$/, ">")
 			pp.text str
 		end
 
 		def clone(*args)
-			method_missing(:clone, *args)
+			Qt::TreeWidgetItem.new(self)
 		end
 
 		def type(*args)
@@ -2331,7 +2420,11 @@ module Qt
 			if metaObject.indexOfSlot(signature) == -1
 				self.class.slots signature
 			end
+			# This makes a difference between being and not being able
+			# to run with GC.stress= true. Reason unclear.
+			gc_disabled= GC.disable
 			@block = block
+			GC.enable unless gc_disabled
 		end
 
 		def invoke(*args)
@@ -2446,58 +2539,52 @@ module Qt
 
 		def Internal.checkarg(argtype, typename)
 			puts "      #{typename} (#{argtype})" if debug_level >= DebugLevel::High
-			if typename =~ /^(const )?QVariant&?$/ && !(typename =~ /QList/)
-				if argtype == "QVariant"
-					return 2
-				else
-					return 0
-				end
-			end
+			const_point = typename =~ /^const\s+/ ? -1 : 0
 			if argtype == 'i'
 				if typename =~ /^int&?$|^signed int&?$|^signed$|^qint32&?$/
-					return 3
+					return 6 + const_point
 				elsif typename =~ /^quint32&?$/
-					return 2
-				elsif typename =~ /^(?:short|ushort|unsigned short int|unsigned short|uchar|uint|long|ulong|unsigned long int|unsigned|float|double|WId|Q_PID|^quint16&?$|^qint16&?$)$/
-					return 2
+					return 4 + const_point
+				elsif typename =~ /^(?:short|ushort|unsigned short int|unsigned short|uchar||unsigned char|uint|long|ulong|unsigned long int|unsigned|float|double|WId|Q_PID|^quint16&?$|^qint16&?$)$/
+					return 4 + const_point
 				elsif typename =~ /^(quint|qint|qulong|qlong|qreal)/
-					return 2
+					return 4 + const_point
 				else 
 					t = typename.sub(/^const\s+/, '')
 					t.sub!(/[&*]$/, '')
 					if isEnum(t)
-						return 1
+						return 2
 					end
 				end
 			elsif argtype == 'n'
 				if typename =~ /^double$|^qreal$/
-					return 3
+					return 6 + const_point
 				elsif typename =~ /^float$/
-					return 2
+					return 4 + const_point
 				elsif typename =~ /^int&?$/
-					return 1
+					return 2 + const_point
 				elsif typename =~ /^(?:short|ushort|uint|long|ulong|signed|unsigned|float|double)$/
-					return 1
+					return 2 + const_point
 				else 
 					t = typename.sub(/^const\s+/, '')
 					t.sub!(/[&*]$/, '')
 					if isEnum(t)
-						return 1
+						return 2 + const_point
 					end
 				end
 			elsif argtype == 'B'
 				if typename =~ /^(?:bool)[*&]?$/
-					return 1
+					return 2 + const_point
 				end
 			elsif argtype == 's'
 				if typename =~ /^(const )?((QChar)[*&]?)$/
-					return 3
-				elsif typename =~ /^(?:u?char\*)$/
-					return 2
-				elsif typename =~ /^(?:const u?char\*)$/
-					return 1
+					return 6 + const_point
+				elsif typename =~ /^(?:(u(nsigned )?)?char\*)$/
+					return 4 + const_point
+				elsif typename =~ /^(?:const (u(nsigned )?)?char\*)$/
+					return 2 + const_point
 				elsif typename =~ /^(?:(?:const )?(QString)[*&]?)$/
-					return 4
+					return 8 + const_point
 				end
 			elsif argtype == 'a'
 				# FIXME: shouldn't be hardcoded. Installed handlers should tell what ruby type they expect.
@@ -2511,35 +2598,35 @@ module Qt
 						    char\*\*
 						)
 					        )$/x
-					return 1
+					return 2 + const_point
 				end
 			elsif argtype == 'u'
 				# Give nil matched against string types a higher score than anything else
 				if typename =~ /^(?:u?char\*|const u?char\*|(?:const )?((Q(C?)String))[*&]?)$/
-					return 2
+					return 4 + const_point
 				# Numerics will give a runtime conversion error, so they fail the match
 				elsif typename =~ /^(?:short|ushort|uint|long|ulong|signed|unsigned|int)$/
 					return -99
 				else
-					return 1
+					return 2 + const_point
 				end
 			elsif argtype == 'U'
 				if typename =~ /QStringList/
-					return 2
+					return 4 + const_point
 				else
-					return 1
+					return 2 + const_point
 				end
 			else
 				t = typename.sub(/^const\s+/, '')
 				t.sub!(/(::)?Ptr$/, '')
 				t.sub!(/[&*]$/, '')
 				if argtype == t
-					return 2
+					return 4 + const_point
 				elsif classIsa(argtype, t)
-					return 1
+					return 2 + const_point
 				elsif isEnum(argtype) and 
 						(t =~ /int|qint32|uint|quint32|long|ulong/ or isEnum(t))
-					return 1
+					return 2 + const_point
 				end
 			end
 			return -99
@@ -2613,12 +2700,7 @@ module Qt
 				elsif arg.kind_of? Array or arg.kind_of? Hash
 					methods.collect! { |meth| meth << '?' }
 				else
-					temp = []
-					methods.collect! do |meth|
-						temp << meth + '#'
-						meth << '$'
-					end
-					methods.concat(temp)
+					methods.collect! { |meth| meth << '$' }
 				end
 			end
 			
@@ -2650,7 +2732,7 @@ module Qt
 				methodIds.each do
 					|id|
 					puts "matching => smoke: #{id.smoke} index: #{id.index}" if debug_level >= DebugLevel::High
-					current_match = 0
+					current_match = (isConstMethod(id) ? 1 : 0)
 					(0...args.length).each do
 						|i|
 						current_match += checkarg(get_value_type(args[i]), get_arg_type_name(id, i))
@@ -2739,19 +2821,6 @@ module Qt
 				getAllParents(c, res)
 			end
 		end
-
-		#
-		# From the enum MethodFlags in qt-copy/src/tools/moc/generator.cpp
-		#
-		AccessPrivate = 0x00
-		AccessProtected = 0x01
-		AccessPublic = 0x02
-		MethodMethod = 0x00
-		MethodSignal = 0x04
-		MethodSlot = 0x08
-		MethodCompatibility = 0x10
-		MethodCloned = 0x20
-		MethodScriptable = 0x40
 	
 		# Keeps a hash of strings against their corresponding offsets
 		# within the qt_meta_stringdata sequence of null terminated
@@ -2805,7 +2874,7 @@ module Qt
 				if dbus
 					data.push MethodScriptable | MethodSignal | AccessPublic
 				else
-					data.push MethodSignal | AccessProtected	# flags, always protected for now
+					data.push entry.access	# flags, always protected for now
 				end
 			end
 
@@ -2817,7 +2886,7 @@ module Qt
 				if dbus
 					data.push MethodScriptable | MethodSlot | AccessPublic	# flags, always public for now
 				else
-					data.push MethodSlot | AccessPublic		# flags, always public for now
+					data.push entry.access		# flags, always public for now
 				end
 			end
 
@@ -2911,7 +2980,7 @@ module Qt
 	#  :full_name is 'foobar(QString,bool)'
 	#  :arg_types is 'QString,bool'
 	#  :reply_type is 'int'
-	QObjectMember = Struct.new :name, :full_name, :arg_types, :reply_type
+	QObjectMember = Struct.new :name, :full_name, :arg_types, :reply_type, :access
 
 	class MetaInfo
 		attr_accessor :classinfos, :dbus, :signals, :slots, :metaobject, :mocargs, :changed
@@ -2928,7 +2997,7 @@ module Qt
 			Internal.addMetaObjectMethods(klass)
 		end
 		
-		def add_signals(signal_list)
+		def add_signals(signal_list, access)
 			signal_names = []
 			signal_list.each do |signal|
 				if signal.kind_of? Symbol
@@ -2936,7 +3005,11 @@ module Qt
 				end
 				signal = Qt::MetaObject.normalizedSignature(signal).to_s
 				if signal =~ /^(([\w,<>:]*)\s+)?([^\s]*)\((.*)\)/
-					@signals.push QObjectMember.new($3, $3 + "(" + $4 + ")", $4, ($2 == 'void' || $2.nil?) ? "" : $2)
+					@signals.push QObjectMember.new(	$3, 
+					                                    $3 + "(" + $4 + ")", 
+					                                    $4, 
+					                                    ($2 == 'void' || $2.nil?) ? "" : $2,
+					                                    access )
 					signal_names << $3
 				else
 					qWarning( "#{@klass.name}: Invalid signal format: '#{signal}'" )
@@ -2959,14 +3032,18 @@ module Qt
 			return all_signals
 		end
 		
-		def add_slots(slot_list)
+		def add_slots(slot_list, access)
 			slot_list.each do |slot|
 				if slot.kind_of? Symbol
 					slot = slot.to_s + "()"
 				end
 				slot = Qt::MetaObject.normalizedSignature(slot).to_s
 				if slot =~ /^(([\w,<>:]*)\s+)?([^\s]*)\((.*)\)/
-					@slots.push QObjectMember.new($3, $3 + "(" + $4 + ")", $4, ($2 == 'void' || $2.nil?) ? "" : $2)
+					@slots.push QObjectMember.new(	$3, 
+					                                $3 + "(" + $4 + ")", 
+					                                $4, 
+					                                ($2 == 'void' || $2.nil?) ? "" : $2,
+					                                access )
 				else
 					qWarning( "#{@klass.name}: Invalid slot format: '#{slot}'" )
 				end
@@ -3029,7 +3106,7 @@ class Object
 	# matter for the intended use in invoking blocks as Qt slots.
 	def instance_exec(*arguments, &block)
 		block.bind(self)[*arguments]
-	end
+	end unless defined? instance_exec
 end
 
 class Proc 
